class Docker0sException(Exception):
    pass


class DefinitionError(Docker0sException):
    """
    A problem in the manifest definition syntax or logic
    """

    pass


class UsageError(Docker0sException):
    """
    A problem in how the app has been used
    """

    pass


class ExecutionError(Docker0sException):
    """
    A problem when trying to perform an action
    """

    pass



"""
GitHub helpers
"""
from __future__ import annotations

import hashlib
import re
import shlex
import subprocess
from functools import lru_cache
from pathlib import Path, PosixPath
from typing import TYPE_CHECKING

from .exceptions import DefinitionError, ExecutionError
from .settings import CACHE_PATH


if TYPE_CHECKING:
    from .host import Host

GIT_SSH_PATTERN = re.compile(
    # url: git@github.com:username/repo
    r"^git\+ssh://(?:(?P<repo>.+?:.+?))"
    # ref: a tag, branch or commit
    r"(@(?P<ref>.+?))?"
    # path: a file within the repo
    r"(#(?P<path>.+?))?"
    # name: the name of the object in the manifest
    r"(::(?P<name>.+?))?$"
)
GIT_HTTPS_PATTERN = re.compile(
    # url: https://github.com/username/repo
    r"^git\+(?P<repo>https://.+?)"
    # ref: a tag, branch or commit
    r"(@(?P<ref>.+?))?"
    # path: a file within the repo
    r"(#(?P<path>.+?))?"
    # name: the name of the object in the manifest
    r"(::(?P<name>.+?))?$"
)
GIT_REMOTE_SHOW_PATTERN = re.compile(r"HEAD branch: (\S+)")


class CommandError(ValueError):
    def __init__(
        self,
        *args,
        cwd: Path | None = None,
        result: subprocess.CompletedProcess | None = None,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.cwd = cwd
        self.result = result

    def __str__(self):
        msg = f"{self.args[0]}\n  cwd={self.cwd}"
        if self.result:
            msg += (
                f"\n  returncode={self.result.returncode}"
                f"\n  stdout={self.result.stdout.decode()}"
                f"\n  stderr={self.result.stderr.decode()}"
            )
        return msg


def call(
    *cmd: str,
    cwd: Path | None = None,
) -> subprocess.CompletedProcess:
    # This specific invocation will allow git to use the system's ssh agent
    result = subprocess.run(
        shlex.join(cmd),
        cwd=cwd,
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        shell=True,
        start_new_session=True,
    )
    return result


def call_or_die(
    *cmd: str,
    cwd: Path | None = None,
    expected: str | None = None,
) -> subprocess.CompletedProcess:
    result = call(*cmd, cwd=cwd)

    if result.returncode != 0:
        raise CommandError(
            f"Command failed with exit code {result.returncode}",
            cwd=cwd,
            result=result,
        )

    if expected and expected not in result.stdout.decode():
        raise CommandError(
            "Command failed with unexpected output",
            cwd=cwd,
            result=result,
        )
    return result


def parse_git_url(path: str) -> tuple[str, str | None, str | None, str | None]:
    """
    Parses a git URL in the formats:

        git+ssh://url@ref#path/to/file::name
        git+https://url@ref#path/to/file::name

    and returns a tuple of (repo, ref, path, name)
    """
    if path.startswith("git+ssh://"):
        pattern = GIT_SSH_PATTERN
    elif path.startswith("git+https://"):
        pattern = GIT_HTTPS_PATTERN
    else:
        # Cannot support
        raise DefinitionError(f"Unrecognised git URL format {path}")

    # Pattern match
    matches = pattern.match(path)
    if not matches:
        raise DefinitionError(f"Unrecognised git URL format {path}")
    data = matches.groupdict()

    return (data["repo"], data["ref"], data["path"], data["name"])


def _parse_remote_show_to_head(raw: str) -> str:
    matches = GIT_REMOTE_SHOW_PATTERN.match(raw)
    if not matches:
        raise ExecutionError(
            'Command "git remote show origin" did not return a HEAD branch'
        )
    return matches.group(1)


@lru_cache()
def fetch_repo(url: str, ref: str | None) -> Path:
    # Build repo path
    repo_dir = hashlib.md5(url.encode()).hexdigest()
    repo_path = CACHE_PATH / repo_dir

    # Clone
    if not repo_path.exists():
        call_or_die("mkdir", "-p", str(repo_path))
        call_or_die("git", "init", cwd=repo_path)
        call_or_die("git", "remote", "add", "origin", url, cwd=repo_path)

    # If no ref, use remote's default branch
    if not ref:
        result = call_or_die(
            "git", "remote", "show", "origin", cwd=repo_path, expected="HEAD branch:"
        )
        ref = _parse_remote_show_to_head(result.stdout.decode())

    # Fetch the ref and check it out
    call_or_die("git", "fetch", "origin", ref, "--depth=1", cwd=repo_path)
    call_or_die("git", "checkout", ref, cwd=repo_path)

    # See if it's a branch or commit
    result = call(
        "git", "rev-parse", "--abbrev-ref", "--verify", f"{ref}@{{u}}", cwd=repo_path
    )
    if result.returncode == 0:
        # It is a branch, use reset to get to head
        call_or_die("git", "reset", "--hard", f"origin/{ref}", cwd=repo_path)

    return repo_path


def fetch_repo_on_host(host: Host, path: PosixPath | str, url: str, ref: str | None):
    # Clone
    if not host.exists(path):
        host.mkdir(path)
        host.exec("git init", cwd=path)
        host.exec("git remote add origin {url}", args={"url": url}, cwd=path)

    # If no ref, use remote's default branch
    if not ref:
        result = host.exec("git remote show origin", cwd=path)
        ref = _parse_remote_show_to_head(result.stdout)

    # Fetch the ref and check it out
    host.exec("git fetch origin {ref} --depth=1", args={"ref": ref}, cwd=path)
    host.exec("git checkout {ref}", args={"ref": ref}, cwd=path)

    # See if it's a branch or commit
    result = host.exec(
        "git rev-parse --abbrev-ref --verify {ref}@{{u}}",
        args={"ref": ref},
        cwd=path,
        can_fail=True,
    )
    if result.ok:
        # It is a branch, use reset to get to head
        host.exec("git reset --hard origin/{ref}", args={"ref": ref}, cwd=path)



from __future__ import annotations

from pathlib import Path

from dotenv import dotenv_values


def read_env(*paths: Path, **values: str | int) -> dict[str, str | int | None]:
    """
    Read env vars from one or more manifest paths and update it with a dict
    """
    env: dict[str, str | int | None] = {}
    for path in paths:
        handle = open(path, "r")
        file_values = dotenv_values(stream=handle)
        env.update(file_values)
        handle.close()

    env.update(values)

    return env


def dump_env(env: dict[str, str | int | None]) -> str:
    """
    Convert an env dict into a multi-line string
    """
    lines = []
    for key, val in env.items():
        if val is None:
            lines.append(key)
        elif isinstance(val, int):
            lines.append(f"{key}={val}")
        else:
            escaped = val.replace('"', r"\"")
            lines.append(f'{key}="{escaped}"')
    return "\n".join(lines)



from functools import update_wrapper
from pathlib import Path

import click

from .app import BaseApp
from .app.names import normalise_name
from .config import Config
from .exceptions import Docker0sException, UsageError
from .manifest import Manifest
from .path import find_manifest


class ExceptionHandlerGroup(click.Group):
    def __call__(self, *args, **kwargs):
        try:
            return self.main(*args, **kwargs)
        except Docker0sException as e:
            click.echo(e, err=True)


@click.group(cls=ExceptionHandlerGroup)
@click.option("--manifest", "-m")
@click.pass_context
def cli(ctx, manifest: str | None = None):
    ctx.ensure_object(dict)

    # Load config
    config_path = Path(click.get_app_dir("docker0s")) / "config.json"
    if config_path.exists():
        config = Config.load(config_path)
    else:
        config = Config(path=config_path)

    # Get manifest path
    manifest_path: Path | None = None
    if manifest:
        manifest_path = Path(manifest)

    elif config.manifest_path:
        manifest_path = Path(config.manifest_path)

    else:
        path_dir = Path.cwd()
        manifest_path = find_manifest(path_dir)
        if manifest_path is None:
            manifest_path = path_dir

    # Try to load manifest
    if manifest_path and manifest_path.is_file():
        manifest_obj = Manifest.load(manifest_path)
    else:
        manifest_obj = None

    ctx.obj.update(
        {
            "config": config,
            "manifest": manifest_obj,
            "manifest_path": manifest_path,
        }
    )


def with_manifest(f):
    @click.pass_context
    def new_func(ctx, *args, **kwargs):
        manifest = ctx.obj["manifest"]
        manifest_path = ctx.obj["manifest_path"]
        if manifest is None:
            raise UsageError(f"Manifest not found at {manifest_path}")
        print(f"Using manifest {manifest_path}")

        return ctx.invoke(f, ctx.obj["manifest"], *args, **kwargs)

    return update_wrapper(new_func, f)


class Target:
    app: str
    service: str | None

    def __init__(self, app: str, service: str | None = None):
        # Normalise app name and ensure the target string is valid
        if not app:
            app = ""
        app_norm = normalise_name(app)

        if not app_norm and service:
            raise UsageError(f"Invalid target .{service} - app missing")
        self.app = app_norm
        self.service = service

    def __str__(self) -> str:
        if self.service:
            return f"{self.app}.{self.service}"
        return self.app


class TargetParamType(click.ParamType):
    name = "target"

    def convert(
        self, value: str, param: click.Parameter | None, ctx: click.core.Context | None
    ) -> Target:
        if ctx is not None and ctx.params.get("all_flag", False):
            return self.fail("Cannot specify both --all and targets")

        # TODO: never seems to reach this bracnh
        elif "." in value:
            parts = value.split(".")
            if len(parts) != 2:
                return self.fail(f"{value!r} is not a app.service target", param, ctx)

            app_name, service_name = parts
        else:
            app_name = value
            service_name = None
        return Target(app=app_name, service=service_name)


class TargetManager:
    manifest: Manifest
    targets = tuple[Target]
    apps: list[BaseApp]
    app_lookup: dict[str, BaseApp]
    service_lookup: dict[BaseApp, list[str]]

    def __init__(self, manifest: Manifest, targets: tuple[Target, ...]):
        self.manifest = manifest
        self.targets = targets

        # Init manifest and prepare lookups
        self.apps = manifest.init_apps(
            *set(target.app for target in targets if target.app)
        )
        self.app_lookup = {app.get_name(): app for app in self.apps}

        # Create service lookup for each app
        self.service_lookup = {}
        for target in targets:
            bound_app: BaseApp = self.app_lookup[target.app]
            if bound_app in self.service_lookup:
                if len(self.service_lookup[bound_app]) == 0:
                    raise UsageError(
                        f"Cannot target mix of app {target.app} and service {target}"
                    )
            else:
                self.service_lookup[bound_app] = []

            if target.service is None:
                continue
            self.service_lookup[bound_app].append(target.service)

    def get_app_services(self):
        for app in self.apps:
            yield (app, self.service_lookup.get(app, []))


TARGET_TYPE = TargetParamType()


@cli.command()
@click.argument("manifest", type=str, required=False)
@click.option("--alias", "-a", type=str)
@click.option("--list", "-l", "list_alias", is_flag=True, default=False)
@click.pass_context
def use(ctx, manifest: str = "", alias: str = "", list_alias: bool = False):
    """
    Set a manifest as the default
    """
    config = ctx.obj["config"]

    if list_alias:
        if not config.manifest_alias:
            print("No aliases defined")
        else:
            print("Available aliases:")
            for alias, path in sorted(config.manifest_alias.items()):
                print(f"  {alias}: {path}")
        return

    if manifest:
        # Load manifest to make sure it works
        manifest_path = Path(manifest).absolute()
        if not manifest_path.is_file():
            # Look for alias
            if manifest in config.manifest_alias:
                manifest_path = Path(config.manifest_alias[manifest])

        manifest = str(manifest_path)
        if not manifest_path.is_file():
            raise UsageError(f"Manifest {manifest} not found")
        Manifest.load(manifest_path)

    # Update config
    if config.manifest_path:
        print(f"Was using manifest {config.manifest_path}")

    # Save new manifest
    config.manifest_path = manifest
    if alias:
        if manifest:
            config.manifest_alias[alias] = manifest
            print(f'Manifest alias "{alias}" saved')
        else:
            config.manifest_alias.pop(alias, None)
            print(f'Manifest alias "{alias}" cleared')
    config.save()

    if manifest:
        print(f"Now using manifest {manifest}")
    else:
        print("Now using no manifest")


@cli.command()
@with_manifest
def ls(manifest: Manifest):
    """
    List all apps

    Usage:
        docker0s ls
    """
    if manifest.host:
        print(f"Host: {manifest.host()}")
    print("Apps:")
    app: type[BaseApp]
    for app in manifest.apps:
        print(f"  {app.get_name()}")


@cli.command()
@with_manifest
@click.argument("apps", type=str, required=False, nargs=-1)
@click.option("--all", "-a", "all_flag", is_flag=True)
def deploy(manifest: Manifest, apps: tuple[str], all_flag: bool = False):
    """
    Deploy one or more apps

    Usage:
        docker0s deploy myapp
        docker0s deploy traefik website
        docker0s deploy --all
    """
    if not apps and not all_flag:
        raise UsageError("Must specify --all or one or more apps")

    safe_apps = (normalise_name(app_name) for app_name in apps)
    bound_apps = manifest.init_apps(*safe_apps)
    for app in bound_apps:
        app.deploy()


@cli.command()
@with_manifest
@click.argument("targets", type=TARGET_TYPE, required=False, nargs=-1)
@click.option("--all", "-a", "all_flag", is_flag=True)
def up(manifest: Manifest, targets: tuple[Target, ...], all_flag: bool = False):
    """
    Bring up all containers for one or more apps or services:
        docker0s up myapp
        docker0s up traefik website.backend

    Bring up all containers for all apps:
        docker0s deploy --all
    """
    if not targets and not all_flag:
        raise UsageError("Must specify --all or one or more targets")

    manager = TargetManager(manifest, targets)
    for app, services in manager.get_app_services():
        app.up(*services)


@cli.command()
@with_manifest
@click.argument("targets", type=TARGET_TYPE, required=False, nargs=-1)
@click.option("--all", "-a", "all_flag", is_flag=True)
def down(manifest: Manifest, targets: tuple[Target, ...], all_flag: bool = False):
    if not targets and not all_flag:
        raise UsageError("Must specify --all or one or more targets")

    manager = TargetManager(manifest, targets)
    for app, services in manager.get_app_services():
        app.down(*services)


@cli.command()
@with_manifest
@click.argument("targets", type=TARGET_TYPE, required=False, nargs=-1)
@click.option("--all", "-a", "all_flag", is_flag=True)
def restart(manifest: Manifest, targets: tuple[Target, ...], all_flag: bool = False):
    if not targets and not all_flag:
        raise UsageError("Must specify --all or one or more targets")

    manager = TargetManager(manifest, targets)
    for app, services in manager.get_app_services():
        app.restart(*services)


@cli.command()
@with_manifest
@click.argument("target", type=TARGET_TYPE)
@click.argument("command", type=str)
def exec(manifest: Manifest, target: Target, command: str):
    if not target.service:
        raise UsageError("Must specify an app.service target")
    app = manifest.init_apps(target.app)[0]
    app.exec(service=target.service, command=command)


@cli.command()
@with_manifest
def status(manifest: Manifest):
    if not manifest.host:
        raise UsageError("No host found in manifest")

    host = manifest.init_host()
    result = host.exec(cmd="docker ps --all", verbose=False)
    print(result.stdout)


@cli.command()
@with_manifest
@click.argument("target", type=TARGET_TYPE)
def logs(manifest: Manifest, target: Target):
    if not target.service:
        raise UsageError("Must specify an app.service target")
    app = manifest.init_apps(target.app)[0]
    app.logs(service=target.service)


@cli.command()
@with_manifest
@click.argument("target", type=Target)
@click.argument("command", type=str)
@click.argument("arguments", nargs=-1, type=str)
def cmd(manifest: Manifest, target: Target, command: str, arguments: list[str]):
    if target.service:
        raise UsageError("Must specify an app target, not an app.service")
    app = manifest.init_apps(target.app)[0]
    cmd_fn = app.get_command(command)
    cmd_fn(*arguments)


def invoke():
    cli(obj={})



from __future__ import annotations

import json
from dataclasses import asdict, dataclass, field
from pathlib import Path
from typing import get_type_hints

from .exceptions import DefinitionError, UsageError


@dataclass
class Config:
    path: Path
    manifest_path: str | None = None
    manifest_alias: dict[str, str] = field(default_factory=dict)

    @classmethod
    def load(cls, path: Path) -> Config:  # TODO: change to Self when supported
        with path.open("r") as file:
            raw_data = json.load(file)

        version = raw_data.pop("version", None)
        if version != 1:
            raise UsageError(f"Invalid config file at {path}")

        annotations = get_type_hints(cls)
        safe_data = {}
        for key, value in raw_data.items():
            if key == "path" or key not in annotations:
                raise DefinitionError(f"Unexpected value {key} for {path}")
            safe_data[key] = value

        safe_data["path"] = path

        return cls(**safe_data)

    def save(self):
        # Prep data
        data = asdict(self)
        del data["path"]
        data["version"] = 1

        # Write
        if not self.path.parent.exists():
            self.path.parent.mkdir()

        with self.path.open("w") as file:
            json.dump(data, file)



from __future__ import annotations

from pathlib import Path
from typing import Any, TypeVar, get_type_hints

from .app.names import normalise_name
from .exceptions import DefinitionError


ManifestObjectType = TypeVar("ManifestObjectType", bound="ManifestObject")


class ManifestObject:
    """
    Base class for manifest objects
    """

    #: Abstract base classes should be marked as abstract so they are ignored by the
    #: manifest loader
    abstract: bool = True

    def __init_subclass__(
        cls, abstract: bool = False, name: str | None = None, **kwargs
    ):
        """
        Set abstract flag and register abstract classes with the registry
        """
        cls.abstract = abstract

        if name is None:
            name = cls.__name__
        else:
            cls.__name__ = name

        # Ensure that names in Python are in CamelCase
        # Names from YAML are normalised by now
        normalised = normalise_name(cls.__name__)
        if normalised != cls.__name__:
            raise DefinitionError(
                f"Python manifest objects must be named in CamelCase: {cls.__name__}"
            )

    @classmethod
    def from_dict(
        cls: type[ManifestObjectType],
        name: str,
        path: Path,
        module: str,
        data: dict[str, Any],
    ) -> type[ManifestObjectType]:
        """
        Build a concrete subclass of this app using the data in the dict

        Args:
            name:   Name of app
            path:   Path of manifest
            module: Name of module
            data:   App attributes
        """
        # No type checking here - see https://github.com/python/mypy/issues/9183 and
        # https://github.com/python/mypy/issues/5865
        class FromDict(cls, name=name, path=path):  # type: ignore
            pass

        FromDict.__module__ = module

        # Collect annotations
        annotations = get_type_hints(FromDict)

        for key, value in data.items():
            if key not in annotations:
                raise DefinitionError(
                    f"Unexpected attribute {key} for {cls.__name__}"
                    f" {FromDict.get_name()} in {FromDict._file}"
                )

            setattr(FromDict, key, value)

        return FromDict



from __future__ import annotations

import inspect
from pathlib import Path, PosixPath
from typing import Any, Callable

from jinja2 import Environment, FileSystemLoader, select_autoescape

from ..env import dump_env, read_env
from ..exceptions import DefinitionError
from ..host import Host
from ..manifest_object import ManifestObject
from ..path import ExtendsPath
from ..settings import DIR_ASSETS, FILENAME_COMPOSE, FILENAME_ENV
from .names import normalise_name, pascal_to_snake


# Abstract app registry for type lookups
abstract_app_registry: dict[str, type[BaseApp]] = {}


class AppsTemplateContext:
    """
    Lazy context getter for use in template context `apps`
    """

    apps: dict[str, BaseApp]

    def __init__(self, apps: dict[str, BaseApp]):
        self.apps = apps

    def __getitem__(self, name: str) -> dict[str, Any]:
        return self.get(name)

    def __getattr__(self, name: str) -> dict[str, Any]:
        return self.get(name)

    def get(self, name: str) -> dict[str, Any]:
        normalised = normalise_name(name)
        if normalised not in self.apps:
            raise DefinitionError(f"Unknown app {name} ({normalised})")
        return self.apps[normalised].get_compose_context()

    def __contains__(self, name: str) -> bool:
        normalised = normalise_name(name)
        return normalised in self.apps


class EnvTemplateContext:
    """
    Lazy context getter for use in template context `env`
    """

    app: BaseApp

    def __init__(self, app: BaseApp):
        self.app = app

    def __getitem__(self, name: str) -> Any:
        return self.get(name)

    def __getattr__(self, name: str) -> Any:
        return self.get(name)

    def get(self, name: str) -> Any:
        env_data = self.app.get_host_env_data()
        return env_data[name]

    def __contains__(self, name: str) -> bool:
        env_data = self.app.get_host_env_data()
        return name in env_data


class BaseApp(ManifestObject, abstract=True):
    _file: Path  # Path to this manifest file
    _dir: Path  # Path to this manifest file

    #: Path to a base docker0s manifest for this app.
    #:
    #: If the path ends ``::<name>`` it will look for an app definition with that name,
    #: eg ``bases.py::Store``. Otherwise it will look for an app with the same
    #: name as this.
    #:
    #: The base manifest must not define a host.
    #:
    #: This referenced manifest will will act as the base manifest. That in turn can
    #: reference an additional base manifest.
    #:
    #: Default: ``d0s-manifest.py``, then ``d0s-manifest.yml``
    extends: str | None = None
    _extends_path: ExtendsPath | None = None  # Resolved path

    #: Path to the app's docker compose file. This will be pushed to the host.
    #:
    #: This can be a ``.yml`` file, or a ``.jinja2`` template.
    #:
    #: For access see ``.get_compose_path``
    #:
    #: Default: ``docker-compose.jinja2``, then ``docker-compose.yml``
    compose: str | None = None

    COMPOSE_DEFAULTS = [
        "docker-compose.j2",
        "docker-compose.jinja2",
        "docker-compose.yml",
        "docker-compose.yaml",
    ]

    #: Context for docker-compose Jinja2 template rendering
    #:
    #: To add instance data, override ``.get_compose_context``
    compose_context: dict[str, Any] | None = None

    #: File containing environment variables for docker-compose
    #:
    #: Path to an env file, or a list of paths
    #:
    #: For access see ``.get_env_data``
    env_file: str | list[str] | None = None

    #: Environment variables for docker-compose
    #:
    #: For access see ``.get_env_data``
    env: dict[str, (str | int)] | None = None

    #: If True, COMPOSE_PROJECT_NAME will be automatically added to the env if not
    #: set by ``env_file`` or ``env``
    set_project_name: bool = True

    # Host this app instance is bound to on initialisation
    host: Host

    # All app instances defined in the manifest which defines this app, including self
    manifest_apps: dict[str, BaseApp]

    def __init_subclass__(
        cls,
        abstract: bool = False,
        name: str | None = None,
        path: Path | None = None,
        **kwargs,
    ):
        """
        Set abstract flag and register abstract classes with the registry
        """
        super().__init_subclass__(abstract=abstract, name=name, **kwargs)

        if abstract:
            global abstract_app_registry  # not required, for clarity
            if cls.__name__ in abstract_app_registry:
                raise DefinitionError(
                    f"Abstract class names must be unique, {cls.__name__} is duplicate"
                )
            abstract_app_registry[cls.__name__] = cls
            return

        # Detect manifest path
        if path:
            cls._file = path
        else:
            cls_module = inspect.getmodule(cls)
            if cls_module is None or cls_module.__file__ is None:
                # This shouldn't happen
                raise ValueError(f"Cannot find module path for app {cls}")
            cls._file = Path(cls_module.__file__)

        cls._dir = cls._file.parent

    def __init__(self, host: Host):
        self.host = host
        self.other_apps: dict[str, BaseApp] = {}

    def __str__(self):
        return self.get_name()

    @classmethod
    def get_name(cls) -> str:
        """
        The docker0s name of this app in PascalCase
        """
        return cls.__name__

    @classmethod
    def get_docker_name(cls) -> str:
        """
        The docker container name of this app in snake_case
        """
        return pascal_to_snake(cls.get_name())

    @classmethod
    def apply_base_manifest(cls, history: list[Path] | None = None):
        """
        If a base manifest can be found by _get_base_manifest, load it and look for a
        BaseApp subclass with the same name as this. If found, add it to the base
        classes for this class.
        """
        # Avoid import loop
        from ..manifest import Manifest

        if not cls.extends:
            return

        if not cls._extends_path:
            cls._extends_path = ExtendsPath(cls.extends, cls._dir)

        path = cls._extends_path.get_manifest()

        base_manifest = Manifest.load(path, history)
        if base_manifest.host is not None:
            raise DefinitionError("A base manifest cannot define a host")

        base_name = cls._extends_path.name or cls.get_name()
        base_app = base_manifest.get_app(base_name)
        if base_app is None:
            raise DefinitionError(
                f"Base manifest {path} does not define an app called {base_name}"
            )

        if not issubclass(cls, base_app):
            cls.__bases__ = (base_app,) + cls.__bases__

    @classmethod
    def find_relative_file(
        cls,
        attr: str,
        defaults: list[str] | None = None,
    ) -> tuple[Path, str]:
        """
        Return first file found, or raises DefinitionError
        """
        cls_values: list[tuple[type[BaseApp], Any]] = cls.collect_attr(attr)
        for mro_cls, val in cls_values:
            if val:
                # Value was specified, file should exist
                path = mro_cls._dir / val
                if not path.exists():
                    raise DefinitionError(
                        f'App setting {cls.get_name()}.{attr} specified as "{val}"'
                        f' but "{path}" does not exist'
                    )
                    break
                return mro_cls._dir, val

            elif defaults:
                # Look for defaults
                for filename in defaults:
                    if (mro_cls._dir / filename).exists():
                        return mro_cls._dir, filename

        raise DefinitionError(
            f"App setting {cls.get_name()}.{attr} not specified, no default found"
        )

    @classmethod
    def find_file(
        cls,
        attr: str,
        defaults: list[str] | None = None,
    ) -> Path:
        path, filename = cls.find_relative_file(attr, defaults)
        return path / filename

    @classmethod
    def collect_attr(cls, attr: str) -> list[tuple[type[BaseApp], Any]]:
        """
        Collect attributes directly from the class and its bases, bypassing inheritance
        and returning a list of ``(cls, *values)`` pairs.

        Abstract classes are ignored.

        This is primarily used where we need context of the value definition - ie a path
        relative to the manifest where the path is set, or for env var resolution.
        """
        # TODO: This is a bit of a confusing approach. Because ``extends`` is adding to
        # the base classes to make it easy for us to inherit custom values and functions
        # from parent apps, we need this odd way to collect values bypassing
        # inheritance. This is unexpected magic.
        #
        # Two better options for a future refactor:
        #
        # 1. Resolve relative paths as soon as the class is defined. This makes sense -
        #    at end of __init_subclass__ we could go through and resolve all the paths
        #    and envs, then we can have normal inheritance working as expected.
        #
        #    The disadvantage is we'll need to resolve everything every time we load a
        #    manifest - and that probably includes resolving env vars. We're generally
        #    not worried about speed in docker0s, but this could easily balloon to be a
        #    problem.
        #
        # 2. Don't inject extended classes as base classes. This may be the better
        #    option - we can keep the JIT evaluation of things we may no need, and base
        #    class creation and injection is verging on too magical for this sort of
        #    project.
        #
        #    Instead we could just create a list of app classes that the host manifest
        #    extends, and iterate through them.  This fn would still exist, but it would
        #    be looking at a ``cls._extend_cls_list`` list instead of ``cls.mro()``.
        #
        #    The disadvantage is we'll lose the ability to reference fns in parent
        #    classes by inheritance - most importantly, deployment hooks. This felt like
        #    it would be a big deal early on, but it seems to be an edge case in the
        #    real world - very few projects have needed custom deployment steps, and I
        #    suspect those could all be handled by actual importing and subclassing.
        results: list[tuple[type[BaseApp], Any]] = []
        for mro_cls in cls.mro():
            if not issubclass(mro_cls, BaseApp) or mro_cls.abstract:
                # We're only interested in concrete App classes
                continue
            results.append((mro_cls, mro_cls.__dict__.get(attr, None)))
        return results

    @classmethod
    def get_compose_path(cls) -> Path:
        return cls.find_file(
            attr="compose",
            defaults=cls.COMPOSE_DEFAULTS,
        )

    @classmethod
    def get_env_data(cls) -> dict[str, str | int | None]:
        """
        Load env files in order (for key conflicts last wins), and then merge in the env
        dict, if defined
        """
        attrs_env: list[tuple[type[BaseApp], Any]] = cls.collect_attr("env")
        attrs_env_file: list[tuple[type[BaseApp], Any]] = cls.collect_attr("env_file")
        attrs = reversed(list(zip(attrs_env, attrs_env_file)))

        env: dict[str, str | int | None] = {}
        env_dict: dict
        env_file: list | str
        for (mro_cls, env_dict), (_, env_file) in attrs:
            # Build list of files
            raw_env_files: list[str] = []
            if env_file is not None:
                if isinstance(env_file, (tuple, list)):
                    raw_env_files = env_file
                else:
                    raw_env_files = [env_file]
            env_files: list[Path] = [
                mro_cls._dir / env_file for env_file in raw_env_files
            ]

            # Prepare dict
            if env_dict is None:
                env_dict = {}

            env.update(read_env(*env_files, **env_dict))

        if cls.set_project_name and "COMPOSE_PROJECT_NAME" not in env:
            env["COMPOSE_PROJECT_NAME"] = cls.get_docker_name()
        return env

    @staticmethod
    def command(fn):
        fn.is_command = True
        return fn

    def get_command(self, name: str) -> Callable:
        """
        Return the specified command
        """
        attr = getattr(self, name)
        if callable(attr) and hasattr(attr, "is_command"):
            return attr
        raise ValueError(f"Command {name} not found")

    @property
    def remote_path(self) -> PosixPath:
        """
        The remote path for this app
        """
        return self.host.path(self.get_docker_name())

    @property
    def remote_compose(self) -> PosixPath:
        """
        A PosixPath to the remote compose file
        """
        return self.remote_path / FILENAME_COMPOSE

    @property
    def remote_env(self) -> PosixPath:
        """
        A PosixPath for the remote env file
        """
        return self.remote_path / FILENAME_ENV

    @property
    def remote_assets(self) -> PosixPath:
        """
        A PosixPath for the remote assets dir

        Assets are resources pushed to the server as part of the docker0s deployment -
        config files, scripts, media etc
        """
        return self.remote_path / DIR_ASSETS

    @property
    def remote_store(self) -> PosixPath:
        """
        A PosixPath for the remote store dir

        The store is for files created by the containers - logs, databases, uploads etc
        """
        return self.remote_path / "store"

    def get_compose_context(self, **kwargs: Any) -> dict[str, Any]:
        """
        Build the template context for the compose template
        """
        context = {
            "host": self.host,
            "env": EnvTemplateContext(self),
            "apps": AppsTemplateContext(self.manifest_apps),
            # Reserved for future expansion
            "docker0s": NotImplemented,
            "globals": NotImplemented,
            **kwargs,
        }

        if self.compose_context is not None:
            context.update(self.compose_context)

        return context

    def get_compose_content(self, context: dict[str, Any] | None = None) -> str:
        """
        Return the content for the docker-compose file

        This will either be rendered from ``compose_template`` if it exists, otherwise
        it will be read from ``compose``
        """
        compose_path = self.get_compose_path()
        filetype = compose_path.suffix.lower()
        if filetype == ".yml":
            return compose_path.read_text()

        elif filetype == ".jinja2":
            env = Environment(
                loader=FileSystemLoader(compose_path.parent),
                autoescape=select_autoescape(),
            )

            context = self.get_compose_context(**(context or {}))
            template = env.get_template(compose_path.name)
            return template.render(context)

        raise ValueError(f"Unrecognised compose filetype {filetype}")

    def get_host_env_data(self) -> dict[str, str | int | None]:
        """
        Build the env data dict to be sent to the server
        """
        env_data = self.get_env_data()
        env_data.update(
            {
                "ENV_FILE": str(self.remote_env),
                "ASSETS_PATH": str(self.remote_assets),
                "STORE_PATH": str(self.remote_store),
            }
        )
        return env_data

    def deploy(self):
        """
        Deploy the env file for this app
        """
        self.push_compose_to_host()
        self.write_env_to_host()
        self.host.ensure_parent_path(self.remote_store)

    def push_compose_to_host(self):
        compose_content: str = self.get_compose_content()
        compose_remote: PosixPath = self.remote_compose
        self.host.write(compose_remote, compose_content)

    def write_env_to_host(self):
        env_dict = self.get_host_env_data()
        env_str = dump_env(env_dict)
        self.host.write(self.remote_env, env_str)

    def call_compose(self, cmd: str, args: dict[str, Any] | None = None):
        """
        Run a docker-compose command on the host
        """
        self.host.call_compose(
            compose=self.remote_compose,
            env=self.remote_env,
            cmd=cmd,
            cmd_args=args,
        )

    def up(self, *services: str):
        """
        Bring up one or more services in this app

        If no services are specified, all services are selected
        """
        if services:
            for service in services:
                self.call_compose("up --build --detach {service}", {"service": service})
        else:
            self.call_compose("up --build --detach")

    def down(self, *services: str):
        """
        Take down one or more containers in this app

        If no services are specified, all services are selected
        """
        if services:
            for service in services:
                self.call_compose(
                    "rm --force --stop -v {service}", {"service": service}
                )
        else:
            self.call_compose("down")

    def restart(self, *services: str):
        """
        Restart one or more services in this app

        If no services are specified, all services are selected
        """
        if services:
            for service in services:
                self.call_compose("restart {service}", {"service": service})
        else:
            self.call_compose("restart")

    def logs(self, service: str):
        """
        Retrieve logs for the given service
        """
        self.call_compose("logs {service}", {"service": service})

    def exec(self, service: str, command: str):
        """
        Execute a command in the specified service

        Command is passed as it arrives, values are not escaped
        """
        self.call_compose(f"exec {{service}} {command}", {"service": service})



from __future__ import annotations

from functools import lru_cache
from io import StringIO
from pathlib import Path, PosixPath
from shlex import quote
from typing import Any

from fabric import Connection
from fabric.runners import Result

from .manifest_object import ManifestObject


class Host(ManifestObject, abstract=True):
    #: Abstract base classes should be marked as abstract so they are ignored by the
    #: manifest loader
    abstract: bool = True

    #: Server hostname
    name: str

    #: Server SSH port
    #:
    #: Default: 22
    port: str | int = 22

    #: Username for login
    user: str | None

    #: Home dir for user
    #:
    #: Use ``{user}`` to replace with the Host.user
    home: str = "/home/{user}/"

    #: Path to the docker0s working dir on the server
    #:
    #: Should be absolute or relative to the connecting user's home directory, do not
    #: use tildes. See fabric docs for details:
    #: https://docs.fabfile.org/en/stable/api/transfer.html
    root_path: str = "apps"

    #: Docker compose command
    #:
    #: Defaults to docker-compose, new installations may prefer ``docker compose``
    compose_command: str = "docker-compose"

    # Internal connection handle
    _connection: Connection | None = None

    def __str__(self) -> str:
        value = self.name
        if self.port:
            value = f"{value}:{self.port}"
        if self.user:
            value = f"{self.user}@{value}"
        return value

    def path(self, app, service: str | None = None) -> PosixPath:
        """
        Remote path builder to ensure consistency
        """
        home = self.home.format(user=self.user)
        path = PosixPath(home) / self.root_path / app
        if service:
            path /= service
        return path

    @property
    def connection(self) -> Connection:
        """
        Create an SSH connection, or retrieve an existing one
        """
        if not self._connection:
            self._connection = Connection(
                host=self.name,
                port=self.port,
                user=self.user,
                connect_kwargs={"allow_agent": True},
                forward_agent=True,
            )
        return self._connection

    def exec(
        self,
        cmd: str,
        args: dict[str, Any] | None = None,
        env: dict[str, Any] | None = None,
        cwd: PosixPath | str | None = None,
        can_fail: bool = False,
        verbose: bool = True,
    ) -> Result:
        """
        Execute a command on the remote server

        Args:
            cmd (str): The command string to execute on the server. Can use named
                placeholders for use with ``.format`` and the ``args`` dict
            args (dict |None): Optional dict of command arguments. These will be escaped
                and passed into cmd.format().
            env (dict | None): Optional dictionary of env vars to run the command with.
                Note that this is an independent dict which will not use App env
                definitions unless you pass env=App.get_env_data()
        """
        if args is not None:
            safe_args = {key: quote(str(val)) for key, val in args.items()}
            cmd = cmd.format(**safe_args)

        result: Result
        with self.connection.cd(str(cwd or "")):
            result = self.connection.run(
                cmd,
                env=env,
                warn=can_fail,
                echo=verbose,
                hide=False if verbose else "both",
                pty=True,
            )

        if not result.ok and not can_fail:
            raise ValueError(f"Command '{cmd}' failed: {result.stderr.strip()}")
        return result

    def call_compose(
        self,
        compose: PosixPath,
        env: PosixPath,
        cmd: str,
        cmd_args: dict[str, Any] | None = None,
    ):
        """
        Execute a docker-compose command on the server

        The paths for compose and env are wrapped in quotes, the command is passed to
        the host unaltered
        """
        args = {"compose": compose, "env": env}
        if cmd_args:
            args.update(cmd_args)
        self.exec(
            cmd=f"{self.compose_command} --file {{compose}} --env-file {{env}} {cmd}",
            args=args,
        )

    def push(self, source: Path, destination: PosixPath):
        """
        Push a file to the server
        """
        self.ensure_parent_path(destination)
        self.connection.put(str(source), str(destination))

    def write(self, filename: PosixPath, content: str):
        """
        Write a file to the server
        """
        self.ensure_parent_path(filename)
        data = StringIO(content)
        self.connection.put(data, str(filename))

    @lru_cache
    def ensure_parent_path(self, filename: PosixPath):
        """
        Given a path to a file, ensure the parent directory exists.

        For example::

            host.ensure_parent_path(root / "app_name" / "env")

        will connect to the server and run::

            mkdir -p "{root}/app_name"

        This will only ever be run once per host.
        """
        self.mkdir(filename.parent)

    @lru_cache
    def mkdir(self, path: PosixPath):
        """
        Make the specified dir and any parent dirs on the host.

        If it already exists, fail silently.

        This will only ever be run once per host.
        """
        self.exec("mkdir -p {path}", args={"path": path})

    def exists(self, path: PosixPath | str) -> bool:
        """
        Return True if the path exists on the host
        """
        result = self.exec("test -e {path}", args={"path": path}, can_fail=True)
        return result.ok



"""
Manifest management
"""
from __future__ import annotations

import sys
from importlib.machinery import ModuleSpec, SourceFileLoader
from importlib.util import module_from_spec
from inspect import isclass
from pathlib import Path

import yaml

from .app import App, BaseApp, abstract_app_registry
from .app.names import normalise_name
from .exceptions import DefinitionError, UsageError
from .host import Host
from .path import path_to_uuid


class Manifest:
    path: Path
    apps: list[type[BaseApp]]
    app_lookup: dict[str, type[BaseApp]]
    host: type[Host] | None = None

    def __init__(self, path: Path):
        self.path = path
        self.apps: list[type[BaseApp]] = []
        self.app_lookup: dict[str, type[BaseApp]] = {}

    def __str__(self) -> str:
        return str(self.path)

    def add_app(self, app: type[BaseApp]) -> None:
        self.apps.append(app)
        self.app_lookup[app.__name__] = app

    def has_app(self, name: str) -> bool:
        return name in self.app_lookup

    def get_app(self, name: str) -> type[BaseApp]:
        if app := self.app_lookup.get(name):
            return app
        raise DefinitionError(f"No app with name {name} in manifest {self.path}")

    def prepare(self, history: list[Path]):
        """
        Prepare apps and host for use

        * Load base manifests for any apps which define an ``extends``
        """
        for app in self.apps:
            app.apply_base_manifest(history=history)

    @classmethod
    def load(cls, path: Path, history: list[Path] | None = None) -> Manifest:
        if not path.exists():
            raise DefinitionError(f"Cannot load {path} - file not found")

        if history is None:
            history = []
        if path in history:
            raise DefinitionError(f"Cannot load {path} - recursive extends detected")
        history.append(path)

        # Load manifest
        filetype = path.suffix.lower()
        if filetype == ".py":
            manifest = cls.load_py(path)
        elif filetype == ".yml":
            manifest = cls.load_yml(path)
        else:
            raise DefinitionError(
                f"Manifest {path} filetype invalid - must be .yml or .py"
            )

        manifest.prepare(history)
        return manifest

    @classmethod
    def load_py(cls, path: Path) -> Manifest:
        # Load module
        module = SourceFileLoader(
            f"docker0s.manifest.loaded.{path_to_uuid(path)}",
            str(path),
        ).load_module()
        setattr(module, "__manifest_path__", path)
        sys.modules[module.__name__] = module

        # Collect apps and hosts
        manifest = Manifest(path)
        for obj in module.__dict__.values():
            if not isclass(obj):
                continue

            if issubclass(obj, BaseApp) and not obj.abstract:
                manifest.add_app(obj)

            elif issubclass(obj, Host) and not obj.abstract:
                if manifest.host is not None:
                    raise ValueError("Cannot define more than one host in a manifest")
                manifest.host = obj

        return manifest

    @classmethod
    def load_yml(cls, path: Path) -> Manifest:
        raw = path.read_text()
        data = yaml.safe_load(raw)

        # Validate top level
        apps_raw = data.pop("apps", [])
        host_raw = data.pop("host", None)
        if len(data) > 0:
            raise DefinitionError(
                f"Error loading {path}: unexpected root elements {', '.join(data.keys())}"
            )
        if not isinstance(apps_raw, dict):
            raise DefinitionError(
                f"Error loading {path}: expecting root apps definition, found {type(apps_raw)}"
            )
        if host_raw and not isinstance(host_raw, dict):
            raise DefinitionError(
                f"Error loading {path}: expecting root host definition, found {type(host_raw)}"
            )

        # Create module and start manifest
        module_spec = ModuleSpec(
            f"docker0s.manifest.loaded.{path_to_uuid(path)}",
            None,
            origin=str(path),
        )
        module = module_from_spec(module_spec)
        module.__file__ = str(path)
        sys.modules[module.__name__] = module
        manifest = Manifest(path)

        # Apps
        for app_name, app_raw in apps_raw.items():
            if not isinstance(app_raw, dict):
                raise DefinitionError(
                    f"Error loading {path}: expecting app definition"
                    f" for {app_name}, found {type(app_raw)}"
                )

            # Get app class
            app_type: str = app_raw.pop("type", None)
            app_base_cls: type[BaseApp]
            if app_type is None:
                app_base_cls = App
            else:
                if app_type not in abstract_app_registry:
                    raise DefinitionError(f"Unknown app type {app_type}")
                app_base_cls = abstract_app_registry[app_type]

            # YAML supports snake case names because it looks nicer.
            # Convert to PascalCase
            name = normalise_name(app_name)
            if manifest.has_app(name):
                raise DefinitionError(
                    f"Error loading {path}: normalised name collision:"
                    f" {app_name} and {name} are equivalent"
                )

            # Build app class and add to manifest
            app_cls: type[BaseApp] = app_base_cls.from_dict(
                name=name, path=path, module=module.__name__, data=app_raw
            )
            setattr(module, name, app_cls)
            manifest.add_app(app_cls)

        # Host
        if host_raw:
            manifest.host = Host.from_dict(
                name="ImportedHost", path=path, module=module.__name__, data=host_raw
            )

        return manifest

    def init_host(self) -> Host:
        """
        Instantiate the host
        """
        if not self.host:
            raise UsageError("Cannot initialise a manifest that has no host")
        return self.host()

    def init_apps(self, *app_names: str) -> list[BaseApp]:
        """
        Given one or more names of apps, find them in the registry and initialise them
        with the manifest host
        """
        # Per-exec host options can be added here later
        host = self.init_host()

        # Find app classes
        app_classes: list[type[BaseApp]]
        if not app_names:
            app_classes = self.apps
        else:
            app_classes = []
            for app_name in app_names:
                app_cls: type[BaseApp] = self.get_app(app_name)
                app_classes.append(app_cls)

        # Initialise the apps
        apps: list[BaseApp] = []
        for app_cls in app_classes:
            apps.append(app_cls(host=host))

        # Tell the apps about each other to allow cross-app logic
        manifest_apps = {a.get_name(): a for a in apps}
        for app in apps:
            app.manifest_apps = manifest_apps

        return apps



import re


sanitise_pattern = re.compile(r"[^A-Za-z0-9_\- ]")
word_pattern = re.compile(r"([A-Z])")


def normalise_name(name: str):
    """
    Normalise camelCase, snake_case and kebab-case to PascalCase

    Numbers are treated as lower-case characters, but cannot start the string
    """
    # Strip anything outside A-Z, 0-9, space, dash or underscore
    norm = sanitise_pattern.sub("", name)

    # Add a space before each capital so we preserve camelCase and PascalCase
    norm = word_pattern.sub(r" \1", norm)

    # Remove snake and kebab joins, title everything and remove spaces
    norm = norm.replace("_", " ").replace("-", " ").title().replace(" ", "")

    # Check it starts with A-Z
    if not word_pattern.match(norm):
        raise ValueError(f"Names must start with A-Z: {name}")

    return norm


def pascal_to_snake(name: str):
    """
    Convert a PascalCase name to snake_case, for use in Docker containers

    Assumed to have come from a normalised_name so no sanitation required
    """
    # Add a _ before each capital and remove the first
    name = word_pattern.sub(r"_\1", name).lstrip("_")

    return name.lower()



from .base import BaseApp


class App(BaseApp, abstract=True):
    """
    A self-contained docker-compose file which deploys containers without additional
    resources
    """

    #: Assets to upload next to the docker-compose.yml
    assets: str | list[str] | None = None

    def deploy(self):
        """
        Deploy the docker-compose and assets for this app
        """
        super().deploy()
        self.push_assets_to_host()

    def push_assets_to_host(self):
        cls_assets = self.collect_attr("assets")
        files: str | list[str]
        for mro_cls, files in cls_assets:
            if not files:
                continue

            if isinstance(files, str):
                files = [files]

            for file in files:
                asset_path = mro_cls._dir / file
                remote_path = self.remote_assets / file
                self.host.push(asset_path, remote_path)



from pathlib import Path, PosixPath

from ..exceptions import DefinitionError
from ..git import fetch_repo_on_host, parse_git_url
from .base import BaseApp


# Remote dir name to clone the repo
REPO_DIR = "repo"


class RepoApp(BaseApp, abstract=True):
    """
    A project which is in a git repository and needs to be cloned to the server and
    mounted into the container as a service

    The ``repo`` must be a ``git+`` repository URL.
    """

    #: ``git+`` URL for the repository to deploy
    repo: str | None = None

    #: Relative path to the compose file within the repository.
    #:
    #: If this path exists, Docker0s will overwrite this file with the ``compose`` file.
    #:
    #: See docs for recommended configuration
    #:
    #: Default: docker-compose.docker0s.yml
    repo_compose: str = "docker-compose.docker0s.yml"

    @classmethod
    def get_repo(cls) -> str:
        """
        Validate repo argument
        """
        if not cls.repo or not cls.repo.startswith(("git+ssh://", "git+https://")):
            raise DefinitionError("RepoApp must set a valid git URL in repo")
        return cls.repo

    @property
    def remote_repo_path(self) -> PosixPath:
        return self.remote_path / REPO_DIR

    @property
    def remote_compose(self) -> PosixPath:
        """
        A PosixPath to the remote compose file
        """
        _: Path
        compose_filename: str = self.repo_compose
        remote_path: PosixPath = self.remote_repo_path / compose_filename
        return remote_path

    def deploy(self):
        self.clone_on_host()
        super().deploy()

    def clone_on_host(self):
        """
        Clone or update the repository on the host
        """
        self.host.ensure_parent_path(self.remote_repo_path)

        # Break up URL into parts
        url, ref, repo_rel_path, name = parse_git_url(self.get_repo())
        if repo_rel_path or name:
            raise DefinitionError(
                f'Invalid setting "{self.get_name()}.repo": Cannot clone a'
                " repository with a relative path or name"
            )
        fetch_repo_on_host(host=self.host, path=self.remote_repo_path, url=url, ref=ref)



from __future__ import annotations

import hashlib
from pathlib import Path
from typing import Any

from .exceptions import DefinitionError
from .git import fetch_repo, parse_git_url


def find_manifest(path: Path) -> Path | None:
    """
    Look within the given Path for a manifest file

    Returns the full Path to the manifest, or None if not found
    """
    files = [
        "d0s-manifest.py",
        "d0s-manifest.yml",
        "d0s-manifest.yaml",
    ]

    # Return the first which exists
    for filename in files:
        filepath = path / filename
        if filepath.exists():
            return filepath

    return None


def path_to_uuid(path: Path) -> str:
    """
    Convert a path into a UUID
    """
    hash = hashlib.md5(str(path).encode()).hexdigest()
    return f"_{hash}"


def path_to_relative(root: Path, path: Path) -> str:
    """
    Given a root path and a sub-path, return the trailing relative path
    """
    if not path.is_relative_to(root):
        raise DefinitionError(f"Path {path} is not a sub-path of {root}")
    relative = str(path)[len(str(root)) :]
    return relative.lstrip("/")


class ExtendsPath:
    """
    Path to a base manifest
    """

    #: Original ``extends`` path, or dir containing a base d0s-manifest
    original: str

    #: Current working directory - directory of the manifest which set ``extends`` - for
    #: resolving relative paths
    cwd: Path

    #: Full ``extends`` path to the manifest, or dir containing a d0s-manifest
    path: Path

    #: Git repository
    repo: str | None = None
    ref: str | None = None

    #: App name within the manifest
    name: str | None = None

    def __init__(self, path: str, cwd: Path):
        """
        Resolve the path to a local Path, retrieving a local copy if a remote source
        """
        self.original = path
        self.cwd = cwd

        if path.startswith(("git+ssh://", "git+https://")):
            # Break up URL into parts
            self.repo, self.ref, repo_rel_path, self.name = parse_git_url(path)

            # Pull and build local path
            repo_local_path = self._pull_repo()
            self.path = (repo_local_path / (repo_rel_path or "")).resolve()

            # Validate local path
            #
            # This is to catch mistakes and bad practice, not security issues - we'll
            # potentially be running Python with no attempt at sandboxing
            if not self.path.is_relative_to(repo_local_path):
                raise DefinitionError(
                    f"Invalid git URL format {path}"
                    f" - repo path {repo_rel_path} not relative to repo root"
                )
        else:
            if "::" in path:
                path, self.name = path.split("::")
            self.path = (self.cwd / path).resolve()

    def __truediv__(self, other: Any) -> Path:
        """
        Add ``other`` to this path, where ``other`` must be within this path
        """
        return (self.path / other).resolve()

    def _pull_repo(self) -> Path:
        """
        Clone local copy of repo
        """
        local_path: Path = fetch_repo(self.repo, self.ref)
        return local_path

    def get_manifest(self) -> Path:
        # If we've been given the path to the manifest file then we're already there
        if self.path.is_file():
            return self.path

        # Not found, we must have a dir to search
        if not self.path.is_dir():
            raise DefinitionError(
                f"Manifest not found at {self.path} ({self.original})"
            )

        # We'll search for these
        filepath = find_manifest(self.path)
        if filepath is None:
            raise DefinitionError(
                f"Manifest not found in {self.path} ({self.original})"
            )
        return filepath



import re
import sys
from pathlib import Path

from setuptools import setup


def find_version(*paths):
    path = Path(*paths)
    content = path.read_text()
    match = re.search(r"^__version__\s*=\s*['\"]([^'\"]*)['\"]", content, re.M)
    if match:
        return match.group(1)
    raise RuntimeError("Unable to find version string.")


# Setup unless this is being imported by Sphinx, which just wants find_version
if "sphinx" not in sys.modules:
    setup(version=find_version("docker0s", "__init__.py"))



from os import getenv
from pathlib import Path


#: Docker0s local path, default ~/.docker0s/
LOCAL_PATH = Path(getenv("DOCKER0S_PATH", "~/.docker0s")).expanduser()

#: Cache dir, default ~/.docker0s/cache/
CACHE_PATH = Path(getenv("DOCKER0S_CACHE_PATH", LOCAL_PATH / "cache")).expanduser()

#: Remote filename for env files
FILENAME_ENV = getenv("DOCKER0S_ENV_FILENAME", "env")

#: Remote filename for compose files
FILENAME_COMPOSE = getenv("DOCKER0S_COMPOSE_FILENAME", "docker-compose.yml")

#: Remote dir to hold assets
DIR_ASSETS = getenv("DOCKER0S_DIR_ASSETS", "assets")
